program Sparse_weight_3d  
  ! Create a sparse weight matrix from a 3D cube.   Jesse Lomask 9/21/04

  use sep
  use lint2
  use warp_mod
  use sorting_mod
  use paint_mod3d

  implicit none

  integer				::nt, nx, ny, non_zero_mask,istatus
  integer				::it,ix,iy,ir,iq,iz,max_dist,dense_dist
  integer                               ::irow,row1,col1x,col1y,row2,col2,val_cnt,n_new_val,rand_jump_int
  integer                               ::box_size,big_box_size,fill
  integer                               ::row2_raw,col2_raw,delta_row2,delta_col2
  integer                               ::row1_grid,col1_grid,row2_grid,col2_grid
  integer                               ::irow_mat,icol_mat,sample_num,nodes
  integer                               ::drow,row1_back,col1_back,row2_back,col2_back
  integer                               ::sparse_mat_index,col_in_row,sparse_index
  integer                               ::box_x,box_t,stat1,dcol,index_new
  integer                               ::total_sparse,row_count,last_index
  integer                               ::newsize,ncv_val,mem_alloc,sample_scale
  integer                               ::col2x,col2y,total_sparse_big,smpl_cnt
  integer                               ::subsam_increment

  real                                  ::maxdat,mindat,rangedat,amp_thresh,zero_tol
  real                                  ::rand_jump,dist,iz_rl, wt,pi_var
  real                                  ::row1_grid_rl,col1_grid_rl,row2_grid_rl,col2_grid_rl

  real,dimension(:)			::topbnd,botbnd,val_list,box_grid_1d,dd,d_vec,d_vec_exp

  real, pointer, dimension(:,:)	        ::sparse_mat,sparse_mat2
  real, pointer, dimension(:,:)	        ::DminusW,DminusWD,DDminusWD
  real, dimension(:,:)                  ::look_up_tab
  integer, dimension(:,:)                  ::sample_template,sample_template2

  real, pointer, dimension(:,:,:)	::data, norm_d,mask_in
  real, dimension(:,:,:)                ::box_grid,box_grid_colx,box_grid_coly,box_grid_row,box_grid_bnds


  ! HOOK UP SEP PARAMETER INPUT FACILITY
  call sep_init(SOURCE)

  ! INPUT PARAMETERS
  call from_history("n1",nt); call from_history("n2",nx);
  call from_history("n3",ny);
  call from_param("max_dist",max_dist,21)
  call from_param("dense_dist",dense_dist,20)
  call from_param("amp_thresh",amp_thresh,.93)
  call from_param("zero_tol",zero_tol,.0001)
  call from_param("ncv_val",ncv_val,20)
  call from_param("mem_alloc",mem_alloc,175000000)
  call from_param("sample_scale",sample_scale,4)
  call from_param("subsam_increment",subsam_increment,1)

  ! ALLOCATE INPUT DATA AND MASK
  allocate(data(nt,nx,ny),mask_in(nt,nx,ny),norm_d(nt,nx,ny))

  ! READ INPUT DATA
  call sep_read(data)

  ! INVERT AND NORMALIZE THE DATA
  data=data*-1.
  maxdat=maxval(data)
  mindat=minval(data)
  rangedat=maxdat-mindat
  norm_d=(data/rangedat-mindat/rangedat)
  total_sparse_big=0
  deallocate(data)

  ! READ IN INPUT MASK
  call sreed("mask_in", mask_in, 4*nt*nx*ny) 

!!$  ! COUNT NONZERO VALUES IN MASK
!!$  non_zero_mask=0
!!$  do it=1,nt
!!$     do ix=1,nx
!!$        do iy=1,ny
!!$           if (mask_in(it,ix,iy)==1.) then
!!$              non_zero_mask=non_zero_mask+1
!!$           end if
!!$        end do
!!$     end do
!!$  end do
!!$  ! MAKE A LOOK UP TABLE TO TRANSFORM DATA TO WEIGHT
!!$  allocate(look_up_tab(non_zero_mask,4))
!!$  ir=0
!!$
!!$  do iy=1,ny
!!$     do ix=1,nx
!!$        !   row_count=0
!!$        do it=1,nt
!!$           if (mask_in(it,ix,iy)==1.) then
!!$              ir=ir+1
!!$              !        row_count=row_count+1
!!$              look_up_tab(ir,1)=ir  ! wt row (or column as they are the same)
!!$              look_up_tab(ir,2)=it !row_count  ! data row
!!$              look_up_tab(ir,3)=ix  ! data column
!!$              look_up_tab(ir,4)=iy  ! data column
!!$           end if
!!$        end do
!!$     end do
!!$  end do
!!$  do it=1,ir
!!$     write(0,*) look_up_tab(it,1:)
!!$  end do



  ! COUNT NONZERO VALUES IN MASK
  non_zero_mask=0

  do iy=1,ny
     do ix=1,nx
        do it=1,nt
           if (norm_d(it,ix,iy)<amp_thresh) then
              mask_in(it,ix,iy)=0.
           end if
           if (real(it)/subsam_increment-floor(real(it)/subsam_increment)<0.001.and.&
           real(ix)/subsam_increment-floor(real(ix)/subsam_increment)<0.001.and.&
           real(iy)/subsam_increment-floor(real(iy)/subsam_increment)<0.001.and.&
           mask_in(it,ix,iy)==1.) then
           non_zero_mask=non_zero_mask+1
        end if
     end do
  end do
end do

! MAKE A LOOK UP TABLE TO TRANSFORM DATA TO WEIGHT
allocate(look_up_tab(non_zero_mask,4))
ir=0
do iy=1,ny
  do ix=1,nx
     !   row_count=0
     do it=1,nt
        if  (real(it)/subsam_increment-floor(real(it)/subsam_increment)<0.001.and.&
        real(ix)/subsam_increment-floor(real(ix)/subsam_increment)<0.001.and.& 
        real(iy)/subsam_increment-floor(real(iy)/subsam_increment)<0.001.and.&
        mask_in(it,ix,iy)==1.) then
           ir=ir+1
           !        row_count=row_count+1
           look_up_tab(ir,1)=ir  ! wt row (or column as they are the same)
           look_up_tab(ir,2)=it  !row_count  ! data row
           look_up_tab(ir,3)=ix  ! data column
           look_up_tab(ir,4)=iy  ! data column
           !    write(0,*) look_up_tab(ir,1:)
        end if
     end do
  end do
end do
!write(0,*) 'non_zero_mask',non_zero_mask

!stop

! ALLOCATE SPARSE MATRIX ...MAKE AS BIG AS ALLOWABLE
allocate(sparse_mat(mem_alloc,3))! 175000000

! ALLOCATE NEW POINT TRACKER AND GRID
box_size=dense_dist*2+1
big_box_size=max_dist*2+1
fill=max_dist-dense_dist
pi_var=acos(-1.)
n_new_val=big_box_size**3 !max number of weights for each node
write(0,*) 'max number of random samples',sample_scale*nint(4.*pi_var*real(fill)) 
allocate(val_list(n_new_val),box_grid(big_box_size,big_box_size,big_box_size),box_grid_row(big_box_size,big_box_size,big_box_size))
allocate(box_grid_colx(big_box_size,big_box_size,big_box_size),d_vec(non_zero_mask),d_vec_exp(non_zero_mask))
allocate(box_grid_coly(big_box_size,big_box_size,big_box_size))
allocate(box_grid_bnds(big_box_size,big_box_size,big_box_size),box_grid_1d(big_box_size*big_box_size*big_box_size))


! MAKE A SAMPLE TEMPLATE
smpl_cnt=0
allocate(sample_template(n_new_val,3))
!max_dist=2.
write(0,*) 'subsampling every ',subsam_increment,'th value'
call create_template3d(sample_template,smpl_cnt,max_dist,subsam_increment)
allocate(sample_template2(smpl_cnt,3))
sample_template2=sample_template(1:smpl_cnt,1:)
deallocate(sample_template)
allocate(sample_template(smpl_cnt,3))
sample_template=sample_template2
deallocate(sample_template2)
write(0,*) 'sample_count',smpl_cnt
!!$do it=1,smpl_cnt
!!$ write(0,*) sample_template(it,1:)
!!$end do
!!$stop


! EXTRACT TOP AND BOTTOM OF MASK
!!$  allocate(topbnd(nx),botbnd(nx))
!!$  topbnd=1
!!$  botbnd=nt
!!$  do ix=1,nx
!!$     do  it=2,nt
!!$        if (mask_in(it,ix)==1..and.mask_in(it-1,ix)==0.) topbnd(ix)=it
!!$        if (mask_in(it,ix)==0..and.mask_in(it-1,ix)==1.) botbnd(ix)=it-1
!!$     end do
!!$  end do


! LOOP OVER EACH ROW OF THE WEIGHT MATRIX
sparse_mat_index=0 ! index of sparse matrix
col_in_row=0       ! keep track of number of column in each row

loop_over_rows:   do irow=1,non_zero_mask

   ! EACH ROW IN THE WEIGHT MATRIX CORRESPONDS TO A POINT IN THE INPUT IMAGE 
   ! HERE WE LOOK UP THE INDEXES OF THAT ROW
  row1=look_up_tab(irow,2)  !assign row
  col1x=look_up_tab(irow,3); !assign col x
  col1y=look_up_tab(irow,4); !assign col y
  ! INITIALIZE GRID PARAMETERS
  box_grid=0.
  box_grid_row=0.
  box_grid_colx=0.
  box_grid_coly=0.
  box_grid_bnds=0.

  ! INITIALIZE NEW POINT TRACKING PARAMETERS
  val_list=0. 
  val_cnt=0.

  ! AROUND EACH POINT IN THE IMAGE, CONTRUCT A GRID TO SAMPLE FROM row1,col1 to row2,col
  ! JUMP RANDOM SHIFTS WHEN ABOVE OR BELOW MASK BOUNDARY
  call  make_grid3d(nx,ny,nt,max_dist,row1,col1x,col1y,row2,col2x,col2y,topbnd,botbnd,box_grid,box_grid_row,box_grid_colx,box_grid_coly,box_grid_bnds,norm_d,mask_in)


  !   box_grid_1d=reshape(box_grid,(/big_box_size*big_box_size*big_box_size/))
  do iy=1,big_box_size
     do ix=1,big_box_size
        box_grid_1d(1+(ix-1)*big_box_size+(iy-1)*big_box_size**2 : big_box_size+(ix-1)*big_box_size+(iy-1)*big_box_size**2 )=box_grid(1:,ix,iy)
     end do
  end do

  ! LOOP OVER GRID CALCULATING WEIGHTS FOR RANDOMLY SAMPLED POINTS
  !  call calc_wt_rand(nx,nt,mem_alloc,sample_scale,dense_dist,max_dist,amp_thresh,zero_tol,row1,col1,box_grid_1d,box_grid,box_grid_row,box_grid_col,box_grid_bnds,sparse_mat,sparse_mat_index,col_in_row,look_up_tab,val_list,val_cnt,norm_d)
  !  call calc_wt_rand3d(nx,ny,nt,mem_alloc,sample_scale,dense_dist,max_dist,amp_thresh,&
  !  zero_tol,row1,col1x,col1y,box_grid_1d,box_grid,box_grid_row,box_grid_colx,box_grid_coly,&
  !  box_grid_bnds,sparse_mat,sparse_mat_index,col_in_row,look_up_tab,val_list,val_cnt,norm_d,total_sparse_big)

  call calc_wt_template3d(sample_template,nx,ny,nt,mem_alloc,sample_scale,dense_dist,max_dist,amp_thresh,&
  zero_tol,row1,col1x,col1y,box_grid_1d,box_grid,box_grid_row,box_grid_colx,box_grid_coly,&
  box_grid_bnds,sparse_mat,sparse_mat_index,col_in_row,look_up_tab,val_list,val_cnt,norm_d,total_sparse_big)

  if (real(irow)/100.-floor(real(irow)/100.)<0.001) then
     write(0,*) 'finished row',irow,'of ',non_zero_mask
  end if
end do loop_over_rows

! RITE OUT ANY REMAINING SPARSE VALUES
write(0,*) 'sparse_mat_index',sparse_mat_index
if (sparse_mat_index>0) then
  sparse_mat_index=sparse_mat_index+1 
  call write_data(sparse_mat,sparse_mat_index,total_sparse_big,mem_alloc) 
end if



call to_history( "n1", 3);call to_history( "n2", total_sparse_big );call to_history( "n3", 1 )
call to_history( "d1", 1.);          call to_history( "d2", 1.);  call to_history( "d3", 1.)
call to_history( "o1", 0.);          call to_history( "o2", 0.);  call to_history( "o3", 0.)
call to_history("esize",4);
! RITE TO THE HISTORY FILE THE SIZE OF THE IMAGE WITHIN THE MASK AND OTHER PARAMETERS
call to_history("nt",nt); call to_history("nx",nx); call to_history("ny",ny)
call to_history("non_zero_mask",non_zero_mask);
call to_history("ncv_val",ncv_val);
deallocate(sparse_mat)





call to_history("n1",non_zero_mask,"look_up"); call to_history("n2",4,"look_up") 
call to_history("n3",1.,"look_up")
call to_history("d1",1.,"look_up"); call to_history("d2",1,"look_up")
call to_history("d3",1.,"look_up")
call to_history("o1",0.,"look_up"); call to_history("o2",0.,"look_up")
call to_history("o3",0.,"look_up")
call to_history("esize",4,"look_up")
call sep_write(look_up_tab,"look_up")
deallocate(look_up_tab)

call to_history("n1",nt,"norm_d"); call to_history("n2",nx,"norm_d") 
call to_history("n3",ny,"norm_d")
call to_history("d1",1.,"norm_d"); call to_history("d2",1,"norm_d")
call to_history("d3",1.,"norm_d")
call to_history("o1",0.,"norm_d"); call to_history("o2",0.,"norm_d")
call to_history("o3",0.,"norm_d")
call to_history("esize",4,"norm_d")
call sep_write(norm_d,"norm_d")
deallocate(norm_d)



end program Sparse_weight_3d

